/*
 * $Id: xwin.c,v 1.4 2004/05/24 22:39:58 shaper Exp $
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "xwin.h"

/* typedefs */
typedef struct {
    Pixmap pixmap;
    Pixmap mask;
    XpmAttributes attributes;
} XpmIcon;

/* X11 Variables */
int x_fd;
XSizeHints mysizehints;
XWMHints mywmhints;
Pixel back_pix, fore_pix;
char* Geometry = "";
XpmIcon wmgen;
Window Root;
int screen; 
int DisplayDepth;

/*
 * Flushes all window events for the supplied window and returns the
 * number of handled window events.
 */
int
xw_flushWindowExposure (Window w)
{
    XEvent dummy;
    int ii = 0;

    while (XCheckTypedWindowEvent(display, w, Expose, &dummy)) {
        ii++;
    }

    return ii;
}

/*
 * Initializes X-Windows.
 */
void
xw_initXwindow (int argc, char* argv[])
{
    int i;
    char* display_name = NULL;

    for (i = 1; argv[i]; ++i) {
        if (!strcmp(argv[i], "-display")) display_name = argv[i+1];
    }

    if (!(display = XOpenDisplay(display_name))) {
        fprintf(stderr, "%s: can't open display %s\n",
                argv[0], XDisplayName(display_name));
        exit(EXIT_FAILURE);
    }

    screen = DefaultScreen(display);
    Root = RootWindow(display, screen);
    DisplayDepth = DefaultDepth(display, screen);
    x_fd = XConnectionNumber(display);
}

/*
 * Creates and set up the main and icon windows.
 */
void
xw_openXwindow (int argc, char* argv[], char* bgcolor, char* fgcolor)
{
    unsigned int borderwidth = 1;
    XClassHint classHint;
    char* wname = argv[0];
    XTextProperty name;
    XGCValues gcv;
    unsigned long gcm;
    int dummy = 0;
    XSizeHints mysizehints;
    Pixel back_pix, fore_pix;
    char** fonts;
    int fontCount;
    int maxFontCount = 20000;

    /* Create a window */
    mysizehints.flags = USSize | USPosition;
    mysizehints.x = 0;
    mysizehints.y = 0;

    back_pix = xw_getColor(bgcolor);
    fore_pix = xw_getColor(fgcolor);

    XWMGeometry(display, screen, Geometry, NULL, borderwidth,
                &mysizehints, &mysizehints.x, &mysizehints.y,
                &mysizehints.width, &mysizehints.height, &dummy);

    mysizehints.width = WINDOW_WIDTH;
    mysizehints.height = WINDOW_HEIGHT;

    win = XCreateSimpleWindow(display, Root, mysizehints.x, mysizehints.y,
                              mysizehints.width, mysizehints.height,
                              borderwidth, fore_pix, back_pix);

    iconwin = XCreateSimpleWindow(display, win, mysizehints.x, mysizehints.y,
                                  mysizehints.width - DISPLAY_BUFFER,
                                  mysizehints.height - DISPLAY_BUFFER,
                                  borderwidth, fore_pix, back_pix);

    /* Activate hints */
    XSetWMNormalHints(display, win, &mysizehints);
    classHint.res_name = wname;
    classHint.res_class = wname;
    XSetClassHint(display, win, &classHint);

    /* Set up the xevents that you want the relevant windows to inherit
     * Currently, its seems that setting KeyPress events here has no
     * effect. I.e. for some you will need to Grab the focus and then
     * return it after you are done... */
    XSelectInput(display, win,
                 ButtonPressMask | ExposureMask | ButtonReleaseMask |
                 PointerMotionMask | StructureNotifyMask | EnterWindowMask |
                 LeaveWindowMask | KeyPressMask | KeyReleaseMask);
    XSelectInput(display, iconwin,
                 ButtonPressMask | ExposureMask | ButtonReleaseMask |
                 PointerMotionMask | StructureNotifyMask | EnterWindowMask |
                 LeaveWindowMask | KeyPressMask | KeyReleaseMask);

    if (XStringListToTextProperty(&wname, 1, &name) == 0) {
        fprintf(stderr, "%s: can't allocate window name\n", wname);
        exit(EXIT_FAILURE);
    }

    XSetWMName(display, win, &name);

    /* create a graphics context for drawing to the window */
    gcm = GCForeground | GCBackground | GCGraphicsExposures;
    gcv.foreground = fore_pix;
    gcv.background = back_pix;
    gcv.graphics_exposures = 0;
    NormalGC = XCreateGC(display, Root, gcm, &gcv);

    /* create the offscreen pixmap for double-buffering */
    offpixmap = XCreatePixmap(
        display, win, DISPLAY_WIDTH, DISPLAY_HEIGHT, DisplayDepth);
    OffscreenGC = XCreateGC(display, offpixmap, gcm, &gcv);
    /* scrutinize the available fonts */
    fonts = XListFonts(display, FONT_PATTERN, maxFontCount, &fontCount);
    if (fonts != NULL) {
        displayFontName = strdup(fonts[0]);
        displayFont = XLoadFont(display, displayFontName);

        /*
        int ii;
        for (ii = 0; ii < fontCount; ii++) {
            fprintf(stderr, "font[%d]: %s\n", ii, fonts[ii]);
        }
        */
    }
    
    /* configure the window hints */
    mywmhints.initial_state = WithdrawnState;
    mywmhints.icon_window = iconwin;
    mywmhints.icon_x = mysizehints.x;
    mywmhints.icon_y = mysizehints.y;
    mywmhints.window_group = win;
    mywmhints.flags =
        StateHint | IconWindowHint | IconPositionHint | WindowGroupHint;
    XSetWMHints(display, win, &mywmhints);

    XSetCommand(display, win, argv, argc);
    XMapWindow(display, win);
}

/*
 * Populates width and height with the dimensions of the given string for
 * the supplied display and graphics context in the current font.
 */
void
xw_getStringDimensions (Display* display, GC gc, char* str,
                        int* width, int* height)
{
    XGCValues gcv;
    XFontStruct* xfs;
    int length = strlen(str);

    XGetGCValues(display, gc, GCFont, &gcv);
    xfs = XQueryFont(display, gcv.font);
/*    fprintf(stderr, "font=%d, xfs=%x\n", gcv.font, xfs); */
    if (xfs == NULL) {
        *width = 0;
        *height = 0;

    } else {
        int direction, fontAscent, fontDescent;
        XCharStruct xcs;

        XTextExtents(xfs, str, length, &direction, &fontAscent,
                     &fontDescent, &xcs);
        *width = xcs.width;
        *height = xcs.ascent + xcs.descent;

        XFreeFontInfo(NULL, xfs, 1);
        xfs = NULL;
    }
}

/*
 * Returns a pixel of the given color.
 */
unsigned long
xw_getColor (char* colorName)
{
    XColor color;
    XWindowAttributes attributes;
    float fac = 1.0;

    XGetWindowAttributes(display, Root, &attributes);
    color.pixel = 0;

    XParseColor(display, attributes.colormap, colorName, &color);
    color.red = (unsigned short)(fac*(color.red-24) + 24);
    color.blue = (unsigned short)(fac*(color.blue-24) + 24);
    color.green = (unsigned short)(fac*(color.green-24) + 24);
    color.flags = DoRed | DoGreen | DoBlue;
    XAllocColor(display, attributes.colormap, &color);

    return color.pixel;
}

/*
 * Sets the foreground color for the supplied display and context.
 */
void
xw_setColor (Display* display, GC gc, unsigned long color)
{
    XGCValues gcv;

    gcv.foreground = color;
    XChangeGC(display, gc, GCForeground, &gcv);
}
